"use strict";
/* jshint newcap:false */

var domInsert = require("../dom-insert");
var defaultCreateOut = require("../createOut");
var getComponentsContext = require("./ComponentsContext").U_;
var componentsUtil = require("@internal/components-util");
var componentLookup = componentsUtil._n_;
var destroyNodeRecursive = componentsUtil._S_;
var EventEmitter = require("events-light");
var RenderResult = require("../RenderResult");
var SubscriptionTracker = require("listener-tracker");
var inherit = require("raptor-util/inherit");
var updateManager = require("./update-manager");
var morphdom = require("../vdom/morphdom");
var eventDelegation = require("./event-delegation");
var domData = require("./dom-data");
var componentsByDOMNode = domData._q_;
var keyedElementsByComponentId = domData._p_;
var CONTEXT_KEY = "__subtree_context__";

var hasOwnProperty = Object.prototype.hasOwnProperty;
var slice = Array.prototype.slice;

var COMPONENT_SUBSCRIBE_TO_OPTIONS;
var NON_COMPONENT_SUBSCRIBE_TO_OPTIONS = {
  addDestroyListener: false
};

var emit = EventEmitter.prototype.emit;
var ELEMENT_NODE = 1;

function removeListener(removeEventListenerHandle) {
  removeEventListenerHandle();
}

function walkFragments(fragment) {
  var node;

  while (fragment) {
    node = fragment.firstChild;

    if (!node) {
      break;
    }

    fragment = node.fragment;
  }

  return node;
}

function handleCustomEventWithMethodListener(component, targetMethodName, args, extraArgs) {
  // Remove the "eventType" argument
  args.push(component);

  if (extraArgs) {
    args = extraArgs.concat(args);
  }

  var targetComponent = componentLookup[component.Y_];
  var targetMethod = typeof targetMethodName === "function" ? targetMethodName : targetComponent[targetMethodName];
  if (!targetMethod) {
    throw Error("Method not found: " + targetMethodName);
  }

  targetMethod.apply(targetComponent, args);
}

function resolveKeyHelper(key, index) {
  return index ? key + "_" + index : key;
}

function resolveComponentIdHelper(component, key, index) {
  return component.id + "-" + resolveKeyHelper(key, index);
}

/**
 * This method is used to process "update_<stateName>" handler functions.
 * If all of the modified state properties have a user provided update handler
 * then a rerender will be bypassed and, instead, the DOM will be updated
 * looping over and invoking the custom update handlers.
 * @return {boolean} Returns true if if the DOM was updated. False, otherwise.
 */
function processUpdateHandlers(component, stateChanges, oldState) {
  var handlerMethod;
  var handlers;

  for (var propName in stateChanges) {
    if (hasOwnProperty.call(stateChanges, propName)) {
      var handlerMethodName = "update_" + propName;

      handlerMethod = component[handlerMethodName];
      if (handlerMethod) {
        (handlers || (handlers = [])).push([propName, handlerMethod]);
      } else {
        // This state change does not have a state handler so return false
        // to force a rerender
        return;
      }
    }
  }

  // If we got here then all of the changed state properties have
  // an update handler or there are no state properties that actually
  // changed.
  if (handlers) {
    // Otherwise, there are handlers for all of the changed properties
    // so apply the updates using those handlers

    handlers.forEach(function (handler) {
      var propertyName = handler[0];
      handlerMethod = handler[1];

      var newValue = stateChanges[propertyName];
      var oldValue = oldState[propertyName];
      handlerMethod.call(component, newValue, oldValue);
    });

    component._y_();
    component._t_();
  }

  return true;
}

function checkInputChanged(existingComponent, oldInput, newInput) {
  if (oldInput != newInput) {
    if (oldInput == null || newInput == null) {
      return true;
    }

    var oldKeys = Object.keys(oldInput);
    var newKeys = Object.keys(newInput);
    var len = oldKeys.length;
    if (len !== newKeys.length) {
      return true;
    }

    for (var i = len; i--;) {
      var key = oldKeys[i];
      if (!(key in newInput && oldInput[key] === newInput[key])) {
        return true;
      }
    }
  }

  return false;
}

var componentProto;

/**
 * Base component type.
 *
 * NOTE: Any methods that are prefixed with an underscore should be considered private!
 */
function Component(id) {
  EventEmitter.call(this);
  this.id = id;
  this.C_ = null;
  this._E_ = null;
  this.ak_ = null;
  this._w_ = null;
  this.Z_ = null; // Used to keep track of bubbling DOM events for components rendered on the server
  this.X_ = null;
  this.Y_ = null;
  this.al_ = null;
  this.R_ = undefined;
  this._x_ = false;
  this.am_ = undefined;
  this.L_ = false;
  this._c_ = false;
  this.an_ = false;
  this.ao_ = false;
  this.E_ = undefined;

  var ssrKeyedElements = keyedElementsByComponentId[id];

  if (ssrKeyedElements) {
    this.N_ = ssrKeyedElements;
    delete keyedElementsByComponentId[id];
  } else {
    this.N_ = {};
  }
}

Component.prototype = componentProto = {
  A_: true,

  subscribeTo: function (target) {
    if (!target) {
      throw TypeError();
    }

    var subscriptions = this.ak_ || (this.ak_ = new SubscriptionTracker());

    var subscribeToOptions = target.A_ ? COMPONENT_SUBSCRIBE_TO_OPTIONS : NON_COMPONENT_SUBSCRIBE_TO_OPTIONS;

    return subscriptions.subscribeTo(target, subscribeToOptions);
  },

  emit: function (eventType) {
    var customEvents = this.X_;
    var target;

    if (customEvents && (target = customEvents[eventType])) {
      var targetMethodName = target[0];
      var isOnce = target[1];
      var extraArgs = target[2];
      var args = slice.call(arguments, 1);

      handleCustomEventWithMethodListener(this, targetMethodName, args, extraArgs);

      if (isOnce) {
        delete customEvents[eventType];
      }
    }

    return emit.apply(this, arguments);
  },
  getElId: function (key, index) {
    if (!key) {
      return this.id;
    }
    return resolveComponentIdHelper(this, key, index);
  },
  getEl: function (key, index) {
    if (key) {
      var resolvedKey = resolveKeyHelper(key, index);
      var keyedElement = this.N_["@" + resolvedKey];

      if (!keyedElement) {
        var keyedComponentRoot = this.N_[resolvedKey];

        if (keyedComponentRoot) {
          return keyedComponentRoot.nodeType === 1 /** Node.ELEMENT_NODE */
          ? keyedComponentRoot : walkFragments(keyedComponentRoot);
        }
      }

      return keyedElement;
    } else {
      return this.el;
    }
  },
  getEls: function (key) {
    key = key + "[]";

    var els = [];
    var i = 0;
    var el;
    while (el = this.getEl(key, i)) {
      els.push(el);
      i++;
    }
    return els;
  },
  getComponent: function (key, index) {
    var rootNode = this.N_[resolveKeyHelper(key, index)];
    if (/\[\]$/.test(key)) {
      rootNode = rootNode && rootNode[Object.keys(rootNode)[0]];
    }
    return rootNode && componentsByDOMNode.get(rootNode);
  },
  getComponents: function (key) {
    var lookup = this.N_[key + "[]"];
    return lookup ? Object.keys(lookup).map(function (key) {
      return componentsByDOMNode.get(lookup[key]);
    }).filter(Boolean) : [];
  },
  destroy: function () {
    if (this.L_) {
      return;
    }

    var root = this._E_;

    this._O_();

    var nodes = root.nodes;

    nodes.forEach(function (node) {
      destroyNodeRecursive(node);

      if (eventDelegation.ap_(node) !== false) {
        node.parentNode.removeChild(node);
      }
    });

    root.detached = true;

    delete componentLookup[this.id];
    this.N_ = {};
  },

  _O_: function () {
    if (this.L_) {
      return;
    }

    this.aq_();
    this.L_ = true;

    componentsByDOMNode.set(this._E_, undefined);

    this._E_ = null;

    // Unsubscribe from all DOM events
    this._v_();

    var subscriptions = this.ak_;
    if (subscriptions) {
      subscriptions.removeAllListeners();
      this.ak_ = null;
    }
  },

  isDestroyed: function () {
    return this.L_;
  },
  get state() {
    return this.C_;
  },
  set state(newState) {
    var state = this.C_;
    if (!state && !newState) {
      return;
    }

    if (!state) {
      state = this.C_ = new this.S_(this);
    }

    state.ar_(newState || {});

    if (state.an_) {
      this.as_();
    }

    if (!newState) {
      this.C_ = null;
    }
  },
  setState: function (name, value) {
    var state = this.C_;

    if (!state) {
      state = this.C_ = new this.S_(this);
    }
    if (typeof name == "object") {
      // Merge in the new state with the old state
      var newState = name;
      for (var k in newState) {
        if (hasOwnProperty.call(newState, k)) {
          state.at_(k, newState[k], true /* ensure:true */);
        }
      }
    } else {
      state.at_(name, value, true /* ensure:true */);
    }
  },

  setStateDirty: function (name, value) {
    var state = this.C_;

    if (arguments.length == 1) {
      value = state[name];
    }

    state.at_(name, value, true /* ensure:true */
    , true /* forceDirty:true */
    );
  },

  replaceState: function (newState) {
    this.C_.ar_(newState);
  },

  get input() {
    return this.R_;
  },
  set input(newInput) {
    if (this.ao_) {
      this.R_ = newInput;
    } else {
      this._l_(newInput);
    }
  },

  _l_: function (newInput, onInput, out) {
    onInput = onInput || this.onInput;
    var updatedInput;

    var oldInput = this.R_;
    this.R_ = undefined;
    this.au_ = out && out[CONTEXT_KEY] || this.au_;

    if (onInput) {
      // We need to set a flag to preview `this.input = foo` inside
      // onInput causing infinite recursion
      this.ao_ = true;
      updatedInput = onInput.call(this, newInput || {}, out);
      this.ao_ = false;
    }

    newInput = this.al_ = updatedInput || newInput;

    if (this.an_ = checkInputChanged(this, oldInput, newInput)) {
      this.as_();
    }

    if (this.R_ === undefined) {
      this.R_ = newInput;
      if (newInput && newInput.$global) {
        this.am_ = newInput.$global;
      }
    }

    return newInput;
  },

  forceUpdate: function () {
    this.an_ = true;
    this.as_();
  },

  as_: function () {
    if (!this._c_) {
      this._c_ = true;
      updateManager.av_(this);
    }
  },

  update: function () {
    if (this.L_ === true || this.aw_ === false) {
      return;
    }

    var input = this.R_;
    var state = this.C_;

    if (this.an_ === false && state !== null && state.an_ === true) {
      if (processUpdateHandlers(this, state.ax_, state.ay_, state)) {
        state.an_ = false;
      }
    }

    if (this.aw_ === true) {
      // The UI component is still dirty after process state handlers
      // then we should rerender

      if (this.shouldUpdate(input, state) !== false) {
        this.az_();
      }
    }

    this._t_();
  },

  get aw_() {
    return this.an_ === true || this.C_ !== null && this.C_.an_ === true;
  },

  _t_: function () {
    this.an_ = false;
    this._c_ = false;
    this.al_ = null;
    var state = this.C_;
    if (state) {
      state._t_();
    }
  },

  shouldUpdate: function () {
    return true;
  },

  az_: function () {
    var self = this;
    var renderer = self.T_;

    if (!renderer) {
      throw TypeError();
    }

    var input = this.al_ || this.R_;

    updateManager.aA_(function () {
      self._F_(input, false).afterInsert(self.E_);
    });

    this._t_();
  },

  _F_: function (input, isHydrate) {
    var host = this.E_;
    var globalData = this.am_;
    var rootNode = this._E_;
    var renderer = this.T_;
    var createOut = renderer.createOut || defaultCreateOut;
    var out = createOut(globalData);
    out.sync();
    out.E_ = this.E_;
    out[CONTEXT_KEY] = this.au_;

    var componentsContext = getComponentsContext(out);
    var globalComponentsContext = componentsContext.r_;
    globalComponentsContext.aB_ = this;
    globalComponentsContext.a__ = isHydrate;

    renderer(input, out);

    var result = new RenderResult(out);

    var targetNode = out.aj_().a_;

    morphdom(rootNode, targetNode, host, componentsContext);

    return result;
  },

  aC_: function () {
    var root = this._E_;
    root.remove();
    return root;
  },

  _v_: function () {
    var eventListenerHandles = this._w_;
    if (eventListenerHandles) {
      eventListenerHandles.forEach(removeListener);
      this._w_ = null;
    }
  },

  get aD_() {
    var state = this.C_;
    return state && state.D_;
  },

  aE_: function (customEvents, scope) {
    var finalCustomEvents = this.X_ = {};
    this.Y_ = scope;

    customEvents.forEach(function (customEvent) {
      var eventType = customEvent[0];
      var targetMethodName = customEvent[1];
      var isOnce = customEvent[2];
      var extraArgs = customEvent[3];

      if (targetMethodName) {
        finalCustomEvents[eventType] = [targetMethodName, isOnce, extraArgs];
      }
    });
  },

  get el() {
    return walkFragments(this._E_);
  },

  get els() {
    return (this._E_ ? this._E_.nodes : []).filter(function (el) {
      return el.nodeType === ELEMENT_NODE;
    });
  },

  aF_: emit,
  aG_(input, out) {
    this.onCreate && this.onCreate(input, out);
    this.aF_("create", input, out);
  },

  aH_(out) {
    this.onRender && this.onRender(out);
    this.aF_("render", out);
  },

  _y_() {
    this.onUpdate && this.onUpdate();
    this.aF_("update");
  },

  _z_() {
    this.onMount && this.onMount();
    this.aF_("mount");
  },

  aq_() {
    this.onDestroy && this.onDestroy();
    this.aF_("destroy");
  }
};

componentProto.elId = componentProto.getElId;
componentProto.aI_ = componentProto.update;
componentProto.aJ_ = componentProto.destroy;

// Add all of the following DOM methods to Component.prototype:
// - appendTo(referenceEl)
// - replace(referenceEl)
// - replaceChildrenOf(referenceEl)
// - insertBefore(referenceEl)
// - insertAfter(referenceEl)
// - prependTo(referenceEl)
domInsert(componentProto, function getEl(component) {
  return component.aC_();
}, function afterInsert(component) {
  return component;
});

inherit(Component, EventEmitter);

module.exports = Component;